

local function heap(rev)
  local self = {}
  local h = {heap = {}, nodes = {}}

  local function getheap()
    return h.heap
  end

  local function push (k, v)
  assert(v ~= nil, "cannot push nil")
  local t = h.nodes
  local h = h.heap
  local n = #h + 1 -- node position in heap array (leaf)
  local p = (n - n % 2) / 2 -- parent position in heap array
  h[n] = k -- insert at a leaf
  t[k] = v
  while n > 1 and t[h[p]] < v do
    h[p], h[n] = h[n], h[p]
    n = p
    p = (n - n % 2) / 2
  end
end

local function pushrev (k, v)
  assert(v ~= nil, "cannot push nil")
  local t = h.nodes
  local h = h.heap
  local n = #h + 1 -- node position in heap array (leaf)
  local p = (n - n % 2) / 2 -- parent position in heap array
  h[n] = k -- insert at a leaf
  t[k] = v
  while n > 1 and t[h[p]] > v do
    h[p], h[n] = h[n], h[p]
    n = p
    p = (n - n % 2) / 2
  end
end

local function pop ()
  local t = h.nodes
  local h = h.heap
  local s = #h
  assert(s > 0, "cannot pop from empty heap")
  local e = h[1] -- min (heap root)
  local r = t[e]
  local v = t[h[s]]
  h[1] = h[s] -- move leaf to root
  h[s] = nil -- remove leaf
  t[e] = nil
  s = s - 1
  local n = 1 -- node position in heap array
  local p = 2 * n -- left sibling position
  if s > p and t[h[p]] < t[h[p + 1]] then
    p = 2 * n + 1 -- right sibling position
  end
  while s >= p and t[h[p]] > v do -- descend heap?
    h[p], h[n] = h[n], h[p]
    n = p
    p = 2 * n
    if s > p and t[h[p]] < t[h[p + 1]] then
      p = 2 * n + 1
    end
  end
  return e, r
end

local function peek()
   local t = h.nodes
  local h = h.heap
  local s = #h
  assert(s > 0, "cannot pop from empty heap")
  local e = h[1] -- min (heap root)
  local r = t[e]
  return e, r
end

local function removerev (k)
  local t = h.nodes
  local h = h.heap
  local s = k
  assert(s > 0, "cannot pop from empty heap")
  local e = h[1] -- min (heap root)
  local r = t[e]
  local v = t[h[#h]]
  t[h[s]]=nil
  h[s] = h[#h]

   -- move leaf to new place
  h[#h] = nil -- remove leaf
  --t[h[s]]=nil
  local n = k -- node position in heap array (leaf)
  local p = (n - n % 2) / 2 -- parent position in heap array
  --h[n] = k -- insert at a leaf
  --t[k] = v
  while n > 1 and t[h[p]] > v do
    h[p], h[n] = h[n], h[p]
    n = p
    p = (n - n % 2) / 2
  end
  local p
  while n<#h do
    local flag = false
    p = 2 * n
    if t[h[p]]~= nil and t[h[p]] < v then
      h[p], h[n] = h[n], h[p]
      n = p
      flag = true
   else
      p = 2 * n + 1
      if t[h[p]]~= nil and t[h[p]] < v then
        h[p], h[n] = h[n], h[p]
        n = p
        flag = true
      end

   end
   if flag == false then break end
  end

  return true
end


local function remove (k)
  local t = h.nodes
  local h = h.heap
  local s = k
  assert(s > 0, "cannot pop from empty heap")
  local e = h[1] -- min (heap root)
  local r = t[e]
  local v = t[h[#h]]
  t[h[s]]=nil
  h[s] = h[#h]

   -- move leaf to new place
  h[#h] = nil -- remove leaf
  --t[h[s]]=nil
  local n = k -- node position in heap array (leaf)
  local p = (n - n % 2) / 2 -- parent position in heap array
  --h[n] = k -- insert at a leaf
  --t[k] = v
  while n > 1 and t[h[p]] < v do
    h[p], h[n] = h[n], h[p]
    n = p
    p = (n - n % 2) / 2
  end
  local p
  while n<#h do
    local flag = false
    p = 2 * n
    if t[h[p]]~= nil and t[h[p]] > v then
      h[p], h[n] = h[n], h[p]
      n = p
      flag = true
   else
      p = 2 * n + 1
      if t[h[p]]~= nil and t[h[p]] > v then
        h[p], h[n] = h[n], h[p]
        n = p
        flag = true
      end

   end
   if flag == false then break end
  end

  return true
end




local function poprev ()
  local t = h.nodes
  local h = h.heap
  local s = #h
  assert(s > 0, "cannot pop from empty heap")
  local e = h[1] -- min (heap root)
  local r = t[e]
  local v = t[h[s]]
  h[1] = h[s] -- move leaf to root
  h[s] = nil -- remove leaf
  t[e] = nil
  s = s - 1
  local n = 1 -- node position in heap array
  local p = 2 * n -- left sibling position
  if s > p and t[h[p]] > t[h[p + 1]] then
    p = 2 * n + 1 -- right sibling position
  end
  while s >= p and t[h[p]] < v do -- descend heap?
    h[p], h[n] = h[n], h[p]
    n = p
    p = 2 * n
    if s > p and t[h[p]] > t[h[p + 1]] then
      p = 2 * n + 1
    end
  end
  return e, r
end

local function findelem(elem)
  for k,v in pairs(h.heap) do
    if v == elem then return k end
  end
end


local function printheap()
  for k,v in pairs(h.heap) do
    print(k,v)
  end
end

local function isempty () return h.heap[1] == nil end
local function toString()
  local s = "current heap is: "
  while not self.isempty() do s=s..tostring(self.pop()).." "

  end
  return s
end
  --if rev is true expose the corresponding pop and push functions
  if rev == true then
    self.pop = poprev
    self.push = pushrev
    self.remove = removerev
  else
    self.pop = pop
    self.push = push
    self.remove = remove
  end
  self.peek = peek
  self.isempty = isempty
  self.toString = toString
  self.findelem=findelem
  self.getheap = getheap

  self.printheap = printheap
  return self
end

-- export to Utils
if not TTP_Utils then TTP_Utils = {} end
TTP_Utils.heap = {}
TTP_Utils.heap.new = heap




